from __future__ import annotations

# standard imports
import os
import json
import logging

# external imports
from cic_types import MetadataPointer
from cic_types.processor import generate_metadata_pointer
from hexathon import strip_0x

# local imports
from cic.contract.base import Data, data_dir
from cic.utils import object_to_str

logg = logging.getLogger(__name__)


class Meta(Data):
    """Serialize and publish metadata for token.

    The token metadata is any mutable data that is not part of the initial token proof, but published simultaneously as the token nonetheless.

    :param path: Path to settings directory
    :type path: str
    :param writer: Writer interface receiving the output of the processor
    :type writer: cic.writers.OutputWriter
    """

    def __init__(
        self, path=".", writer=None, name="", location="", country_code="KE", contact={}, interactive=False
    ):
        super(Meta, self).__init__()
        self.name = name
        self.contact = contact
        self.country_code = country_code
        self.location = location
        self.path = path
        self.writer = writer
        self.meta_path = os.path.join(self.path, "meta.json")

        if interactive:
            self.name = input(f"Enter Metadata Name ({self.name}): ") or self.name
            self.country_code = input(f"Enter Metadata Country Code ({self.country_code}): ") or self.country_code
            self.location = input(f"Enter Metadata Location ({self.location}): ") or self.location
            
            adding_contact_info = True
            contact = {}
            while adding_contact_info:
                value = input("Enter Metadata contact info (e.g 'phone: +254723522718'): ") or None
                if value:
                    data = value.split(":")
                    if len(data) != 2:
                        print("Invalid contact info, you must enter in the format 'key: value'")
                        continue
                    contact[data[0].strip()] = data[1].strip()
                else:
                    adding_contact_info = False
            self.contact = contact

    def load(self):
        """Load metadata from settings."""
        super(Meta, self).load()

        f = open(self.meta_path, "r", encoding="utf-8")
        o = json.load(f)
        f.close()

        self.name = o["name"]
        self.contact = o["contact"]
        self.country_code = o["country_code"]
        self.location = o["location"]
        self.inited = True

    def start(self):
        """Initialize metadata settings from template."""
        super(Meta, self).start()

        meta_template_file_path = os.path.join(
            data_dir, f"meta_template_v{self.version()}.json"
        )

        f = open(meta_template_file_path, encoding="utf-8")
        o = json.load(f)
        f.close()

        o["name"] = self.name
        o["contact"] = self.contact
        o["country_code"] = self.country_code
        o["location"] = self.location

        f = open(self.meta_path, "w", encoding="utf-8")
        json.dump(o, f, sort_keys=True, indent="\t")
        f.close()

    def reference(self, token_address):
        """Calculate the mutable reference for the token metadata."""
        token_address_bytes = bytes.fromhex(strip_0x(token_address))
        return generate_metadata_pointer(
            token_address_bytes, MetadataPointer.TOKEN_META
        )

    def asdict(self):
        """Output proof state to dict."""
        return {
            "name": self.name,
            "country_code": self.country_code,
            "location": self.location,
            "contact": self.contact,
        }

    def process(self, token_address=None, token_symbol=None, writer=None):
        """Serialize and publish metadata.

        See cic.processor.Processor.process
        """
        if writer is None:
            writer = self.writer

        v = json.dumps(self.asdict(), separators=(",", ":"))

        token_address_bytes = bytes.fromhex(strip_0x(token_address))
        k = generate_metadata_pointer(token_address_bytes, MetadataPointer.TOKEN_META)
        writer.write(k, v.encode("utf-8"))

        token_symbol_bytes = token_symbol.encode("utf-8")
        k = generate_metadata_pointer(
            token_symbol_bytes, MetadataPointer.TOKEN_META_SYMBOL
        )
        writer.write(k, v.encode("utf-8"))

        return (k, v)

    def __str__(self):
        return object_to_str(self, ["name", "contact", "country_code", "location"])




# standard imports
import logging
import os

# local imports
from cic.contract.components.proof import Proof
from cic.contract.components.meta import Meta
from cic.contract.components.attachment import Attachment
from cic.contract.network import Network
from cic.contract.components.token import Token

logg = logging.getLogger(__name__)


def process_args(argparser):
    argparser.add_argument(
        "--target",
        action="append",
        type=str,
        default=[],
        help="initialize network specification file with target",
    )
    argparser.add_argument("--name", type=str, help="token name")
    argparser.add_argument("--symbol", type=str, help="token symbol")
    argparser.add_argument("--precision", type=str, help="token unit precision")
    argparser.add_argument("directory", help="directory to initialize")


def validate_args(args):
    pass


def execute(config, eargs):
    logg.info("initializing in {}".format(eargs.directory))

    os.makedirs(eargs.directory)

    ct = Token(
        eargs.directory, name=eargs.name, symbol=eargs.symbol, precision=eargs.precision
    )
    cp = Proof(eargs.directory)
    cm = Meta(eargs.directory)
    ca = Attachment(eargs.directory)
    cn = Network(eargs.directory, targets=eargs.target)

    ct.start()
    cp.start()
    cm.start()
    ca.start()
    cn.start()



# standard imports
import importlib
import logging
import os
from typing import Optional

# external imports
from cic_types.ext.metadata import MetadataRequestsHandler
from cic_types.ext.metadata.signer import Signer as MetadataSigner

# local imports
from cic.contract.processor import ContractProcessor
from cic.contract.components.proof import Proof
from cic.contract.components.attachment import Attachment
from cic.contract.components.meta import Meta
from cic.contract.network import Network
from cic.contract.components.token import Token
from cic.writers import HTTPWriter, KeyedWriterFactory, MetadataWriter

logg = logging.getLogger(__name__)


def process_args(argparser):
    argparser.add_argument(
        "-d", "--directory", type=str, dest="directory", default=".", help="directory"
    )
    argparser.add_argument(
        "-o",
        "--output-directory",
        type=str,
        dest="output_directory",
        help="output directory",
    )
    argparser.add_argument(
        "--metadata-endpoint",
        dest="metadata_endpoint",
        type=str,
        help="metadata endpoint to interact with",
    )
    argparser.add_argument(
        "-y",
        "--signer",
        type=str,
        dest="y",
        help="target-specific signer to use for export",
    )
    argparser.add_argument("-p", type=str, help="RPC endpoint")
    argparser.add_argument("target", type=str, help="target network type")


def validate_args(args):
    pass


def init_writers_from_config(config):
    w = {
        "meta": None,
        "attachment": None,
        "proof": None,
        "ext": None,
    }
    for v in w.keys():
        k = "CIC_CORE_{}_WRITER".format(v.upper())
        (d, c) = config.get(k).rsplit(".", maxsplit=1)
        m = importlib.import_module(d)
        o = getattr(m, c)
        w[v] = o

    return w


ExtraArgs = {
    "target": str,
    "directory": str,
    "output_directory": str,
    "metadata_endpoint": Optional[str],
    "y": str,
}


def execute(config, eargs: ExtraArgs):
    modname = f"cic.ext.{eargs.target}"
    cmd_mod = importlib.import_module(modname)

    writers = init_writers_from_config(config)

    output_writer_path_meta = eargs.output_directory
    if eargs.metadata_endpoint != None:
        MetadataRequestsHandler.base_url = eargs.metadata_endpoint
        MetadataSigner.gpg_path = os.path.join("/tmp")
        MetadataSigner.key_file_path = config.get("AUTH_KEYFILE_PATH")
        MetadataSigner.gpg_passphrase = config.get("AUTH_PASSPHRASE")
        writers["proof"] = KeyedWriterFactory(MetadataWriter, HTTPWriter).new
        writers["attachment"] = KeyedWriterFactory(None, HTTPWriter).new
        writers["meta"] = MetadataWriter
        output_writer_path_meta = eargs.metadata_endpoint

    ct = Token(path=eargs.directory)
    cm = Meta(
        path=eargs.directory, writer=writers["meta"](path=output_writer_path_meta)
    )
    ca = Attachment(
        path=eargs.directory, writer=writers["attachment"](path=output_writer_path_meta)
    )
    cp = Proof(
        path=eargs.directory,
        attachments=ca,
        writer=writers["proof"](path=output_writer_path_meta),
    )
    cn = Network(path=eargs.directory)

    ca.load()
    ct.load()
    cp.load()
    cm.load()
    cn.load()

    chain_spec = None
    try:
        chain_spec = config.get("CHAIN_SPEC")
    except KeyError:
        chain_spec = cn.chain_spec
        config.add(chain_spec, "CHAIN_SPEC", exists_ok=True)
        logg.debug(f"CHAIN_SPEC config set to {str(chain_spec)}")

    # signer = cmd_mod.parse_signer(eargs.y)
    (rpc, signer) = cmd_mod.parse_adapter(config, eargs.y)

    ref = cn.resource(eargs.target)
    chain_spec = cn.chain_spec(eargs.target)
    logg.debug(
        f"found reference {ref['contents']} chain spec {chain_spec} for target {eargs.target}"
    )
    c = getattr(cmd_mod, "new")(
        chain_spec,
        ref["contents"],
        cp,
        signer_hint=signer,
        rpc=rpc,
        outputs_writer=writers["ext"](path=eargs.output_directory),
    )
    c.apply_token(ct)

    p = ContractProcessor(proof=cp, attachment=ca, metadata=cm, extensions=[c])
    p.process()



def object_to_str(obj, keys):
    """Return a string representation of an object."""
    s = ""
    for key in keys:
        value = eval("obj." + key)
        key = key.replace("()", "")
        if type(value) == str:
            s += f"{key} = {value}\n"
        elif type(value) == list:
            for idx, vv in enumerate(value):
                if not vv:
                    s += f"{key}[{idx}] = \n"
                    continue
                s += f"{key}[{idx}] = {vv}\n"
        elif type(value) == dict:
            for vv_key in value.keys():
                vv_value = value[vv_key]
                if not vv_value:
                    s += f"{key}.{vv_key} = \n"
                    continue
                s += f"{key}.{vv_key} = {vv_value}\n"
        else:
            s += f"{key} = {str(value)}\n"
    return s



# standard imports
import os
import hashlib


mod_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..')
root_dir = os.path.join(mod_dir, '..')
data_dir = os.path.join(mod_dir, 'data')
schema_dir = os.path.join(mod_dir, 'schema')


class Data:
    """Base class for all parts of the token data deployment.
    """

    __default_version = 0

    def __init__(self):
        self.dirty = False
        self.inited = False
        self.__version = self.__default_version
        self.__hasher = self.__basehasher


    def __basehasher(self, v):
        h = hashlib.sha256()
        h.update(v)
        return h.digest()


    def hash(self, v):
        """Compute digest of the given data

        :param v: Data to hash
        :type v: bytes
        :rtype: bytes
        :return: Hashed data
        """
        return self.__hasher(v)


    def load(self):
        """Prevents overwriting data from settings if data state has changed.

        :raises RuntimeError: If state is dirty
        """
        if self.dirty:
            raise RuntimeError('Object contains uncommitted changes')


    def start(self):
        """Prevents double initialization of data item.

        :raises RuntimeError: If already initialized
        """
        if self.inited:
            raise RuntimeError('Object already initialized')


    def verify(self):
        """Verify data state (noop)
        """
        return True


    def version(self):
        """Return version of data schema.
        """
        return self.__version


    def set_version(self, version):
        """Set version of data schema. The version number is a single integer.

        :param version: version
        :type version: int
        """
        self.__version = version



# standard imports
import importlib

# external imports
from chainlib.chain import ChainSpec

# local imports
from cic.contract.network import Network


def process_args(argparser):
    argparser.add_argument(
        "--registry", type=str, help="contract registry address"
    )
    argparser.add_argument(
        "-d", "--directory", type=str, dest="directory", default=".", help="directory"
    )
    argparser.add_argument("-p", type=str, help="RPC endpoint")
    argparser.add_argument("-i", type=str, help="chain spec string")
    argparser.add_argument("target", help="target to initialize")


def validate_args(args):
    pass


def execute(config, eargs):
    cn = Network(eargs.directory, targets=eargs.target)
    cn.load()

    chain_spec = ChainSpec.from_chain_str(eargs.i or config.get("CHAIN_SPEC"))
    m = importlib.import_module(f"cic.ext.{eargs.target}.start")
    m.extension_start(
        cn,
        registry_address=eargs.registry or config.get("CIC_REGISTRY_ADDRESS"),
        chain_spec=chain_spec,
        rpc_provider=config.get("RPC_PROVIDER"),
    ) # TODO add key account address



# standard imports
import json
import logging
import os

# external imports
from chainlib.chain import ChainSpec
# local imports
from cic.contract.base import Data, data_dir

logg = logging.getLogger(__name__)


class Network(Data):
    """Contains network settings for token deployments across extensions.

    Extension targets are defined by the keys immediately following the "resources" key in the network settings file.

    :param path: Path to settings directory
    :type path: str
    :param targets: Extension targets to execute
    :type targets: list of str
    """
    def __init__(self, path='.', targets=[]):
        super(Network, self).__init__()
        self.resources = None
        self.path = path
        self.targets = targets
        self.network_path = os.path.join(self.path, 'network.json')


    def load(self):
        """Load network settings from file.
        """
        super(Network, self).load()

        with open(self.network_path, 'r', encoding='utf-8') as f:
            o = json.load(f)

        self.resources = o['resources']

        self.inited = True


    def start(self):
        """Initialize network settings with targets chosen at object instantiation.

        Will save to network settings file.
        """
        super(Network, self).load()

        network_template_file_path = os.path.join(data_dir, f'network_template_v{self.version()}.json')
        
        with open(network_template_file_path, encoding='utf-8') as f:
            o_part = json.load(f)

        self.resources = {}
        for v in self.targets:
            self.resources[v] = o_part

        self.save()


    def save(self):
        """Save network settings to file.
        """
        with open(self.network_path, 'w', encoding='utf-8') as f:
            json.dump({
                'resources': self.resources,
                }, f, sort_keys=True, indent="\t")


    def resource(self, k):
        """Get settings definitions for a given extension.

        :param k: Extension key
        :type k: str
        :rtype: dict
        :return: Extension settings
        """
        v = self.resources.get(k)
        if v is None:
            raise AttributeError(f'No defined reference for {k}')
        return v


    def resource_set(self, resource_key, content_key, reference, key_account=None):
        """Set the values a content part of an extension setting.
        
        The content parts define network application resources. Each entry is keyed by the name of the application. Each value consists of a key_account used to write/deploy to the contract, and the reference (address) of the application resource. If no application resource yet exists on the network for the part, the reference value will be None.

        :param resource_key: Extension key
        :type resource_key: str
        :param content_key: Resource name (e.g. smart contract name)
        :type content_key: str
        :param reference: Reference to resource on network (e.g. smart contract address)
        :type reference: str
        :param key_account: Address of account to sign transaction for the resource with
        :type key_account: str

        """
        self.resources[resource_key]['contents'][content_key]['reference'] = reference
        self.resources[resource_key]['contents'][content_key]['key_account'] = key_account


    def chain_spec(self, k):
        """Retrieve chain spec for the given extension

        :param k: Extension key
        :type k: str
        :rtype: chainlib.chain.ChainSpec
        :return: Chain spec object
        """
        v = self.resource(k)
        return ChainSpec.from_dict(v['chain_spec'])


    def set(self, resource_key, chain_spec):
        """Set chain spec for resource.

        :param resource_key: Extension key
        :type resource_key: str
        :param chain_spec: Chain spec to set
        :type chain_spec: chainlib.chain.ChainSpec
        """
        chain_spec_dict = chain_spec.asdict()
        for k in chain_spec_dict.keys():
            logg.debug(f'resources: {self.resources}')
            self.resources[resource_key]['chain_spec'][k] = chain_spec_dict[k]


    def __str__(self):
        s = ''
        for resource in self.resources.keys():
            chainspec = ChainSpec.from_dict(self.resources[resource]['chain_spec'])
            s += f'{resource}.chain_spec: {str(chainspec)}\n'
            for content_key in self.resources[resource]['contents'].keys():
                content_value = self.resources[resource]['contents'][content_key]
                if content_value is None:
                    content_value = ''
                s += f'{resource}.contents.{content_key} = {json.dumps(content_value, indent=4, sort_keys=True)}\n'

        return s



# standard imports
import os
import logging
import sys
import json
import requests
import importlib
import tempfile
import hashlib

# local imports
from cic.writers import OutputWriter

log = logging.getLogger(__name__)


CONTRACTS = [
    {
        "url": "https://gitlab.com/cicnet/eth-erc20/-/raw/master/python/giftable_erc20_token/data/GiftableToken",
        "name": "Giftable Token",
    },
    {
        "url": "https://gitlab.com/cicnet/erc20-demurrage-token/-/raw/master/python/erc20_demurrage_token/data/DemurrageTokenSingleNocap",
        "name": "Demurrage Token Single No Cap",
    }
]

# Download File from Url
def download_file(url: str, filename=None) -> (str, bytes):
    directory = tempfile.gettempdir()
    filename = filename if filename else url.split("/")[-1]
    log.debug(f"Downloading {filename}")
    r = requests.get(url, allow_redirects=True)
    content_hash = hashlib.md5(r.content).hexdigest()
    path = os.path.join(directory, content_hash)
    with open(path, "wb") as f:
        f.write(r.content)
    log.debug(f"{filename} downloaded to {path}")
    return path

def get_contract_args(data: list):
    for item in data:
        if item["type"] == "constructor":
            return item["inputs"]
    raise Exception("No constructor found in contract")

def select_contract():
    print("Contracts:")
    print("\t C - Custom (path/url to contract)")
    for idx, contract in enumerate(CONTRACTS):
        print(f"\t {idx} - {contract['name']}")

    val = input("Select contract (C,0,1..): ")
    if val.isdigit() and int(val) < len(CONTRACTS):
        contract = CONTRACTS[int(val)]
        bin_path = os.path.abspath(download_file(contract["url"] + ".bin"))
        json_path = download_file(contract["url"] + ".json")

    elif val == "C":
        possible_bin_location = input("Enter a path or url to a contract.bin: ")
        if possible_bin_location.startswith('http'):
            # possible_bin_location is url
            bin_path = download_file(possible_bin_location)
        else:
            # possible_bin_location is path
            if os.path.exists(possible_bin_location):
                bin_path = os.path.abspath(possible_bin_location)
            else:
                raise Exception(f"File {possible_bin_location} does not exist")

            possible_json_path = val.replace(".bin", ".json")
            if os.path.exists(possible_json_path):
                json_path = possible_json_path
    else:
        print("Invalid selection")
        sys.exit(1)
    contract_extra_args = []
    contract_extra_args_types = []

    if os.path.exists(json_path):
        with open(json_path, encoding="utf-8") as f:
            json_data = json.load(f)
        for contract_arg in get_contract_args(json_data):
            arg_name = contract_arg.get("name")
            arg_type = contract_arg.get("type")
            if arg_name not in ["_decimals", "_name", "_symbol"]:
                val = input(f"Enter value for {arg_name} ({arg_type}): ")
                contract_extra_args.append(val)
                if arg_type == "uint128":
                    contract_extra_args_types.append("uint256")
                else:
                    contract_extra_args_types.append(arg_type)

    return {
        "bin_path": bin_path,
        "json_path": json_path,
        "extra_args": contract_extra_args,
        "extra_args_types": contract_extra_args_types,
    }


Writers = {
    "meta": OutputWriter,
    "attachment": OutputWriter,
    "proof": OutputWriter,
    "ext": OutputWriter,
}

def init_writers_from_config(config) -> Writers:
    writers: Writers = {
        "meta": None,
        "attachment": None,
        "proof": None,
        "ext": None,
    }
    for key in writers:
        writer_config_name = f"CIC_CORE_{key.upper()}_WRITER"
        (module_name, attribute_name) = config.get(writer_config_name).rsplit(".", maxsplit=1)
        mod = importlib.import_module(module_name)
        writer = getattr(mod, attribute_name)
        writers[key] = writer

    return writers



# Standard
import importlib
import json
import logging
import os
from typing import TYPE_CHECKING, List

import requests
from chainlib.chain import ChainSpec
from chainlib.cli.config import Config
from cic.contract.components.attachment import Attachment
from cic.contract.components.meta import Meta
from cic.contract.components.proof import Proof
from cic.contract.components.token import Token
from cic.contract.helpers import init_writers_from_config
from cic.contract.network import Network

# Local Modules
from cic.contract.processor import ContractProcessor
from cic.writers import HTTPWriter, KeyedWriterFactory, MetadataWriter, OutputWriter

# external imports
from cic_types.ext.metadata import MetadataRequestsHandler
from cic_types.ext.metadata.signer import Signer as MetadataSigner

log = logging.getLogger(__name__)


class Contract:
    """ """

    def __init__(
        self,
        token: Token,
        proof: Proof,
        meta: Meta,
        attachment: Attachment,
        network: Network,
    ):
        self.token = token
        self.proof = proof
        self.meta = meta
        self.attachment = attachment
        self.network = network

    def __str__(self):
        s = ""
        s += f"\n[cic.header]\nversion = {self.proof.version()}\n\n"
        s += f"[cic.token]\n{self.token}\n"
        s += f"[cic.proof]\n{self.proof}\n"
        s += f"[cic.meta]\n{self.meta}\n"
        s += f"[cic.attachment]\n{self.attachment}\n"
        s += f"[cic.network]\n{self.network}\n"
        return s


def load_contract(directory) -> Contract:
    token = Token(path=directory)
    proof = Proof(path=directory)
    meta = Meta(path=directory)
    attachment = Attachment(path=directory)
    network = Network(directory)

    token.load()
    proof.load()
    meta.load()
    attachment.load()
    network.load()
    return Contract(
        token=token, proof=proof, meta=meta, attachment=attachment, network=network
    )


def generate_contract(
    directory: str, targets: List[str], config, interactive=True
) -> Contract:
    if os.path.exists(directory):
        contine = input(
            f"Directory {directory} already exists, Would you like to delete it? (y/n): "
        )
        if contine.lower() != "y":
            log.debug("Trying to load existing contract")
            return load_contract(directory)
        else:
            print(f"Deleted {directory}")
            os.system(f"rm -rf {directory}")
    os.makedirs(directory)
    log.debug("Generating token")
    token = Token(directory, interactive=interactive)
    token.start()

    log.debug("Generating proof")
    proof = Proof(directory, interactive=interactive)
    proof.start()

    log.debug("Generating meta")
    meta = Meta(directory, interactive=interactive)
    meta.start()

    log.debug("Generating attachment")
    attachment = Attachment(directory, interactive=interactive)

    log.debug("Generating network")
    network = Network(directory, targets=targets)
    network.start()

    log.debug(
        f"""Populating infomation from network:
    CIC_REGISTRY_ADDRESS: {config.get("CIC_REGISTRY_ADDRESS")}
    CHAIN_SPEC: {config.get("CHAIN_SPEC")}
    RPC_PROVIDER: {config.get("RPC_PROVIDER")}
    """
    )
    for target in targets:
        # TODO Clean this up
        modname = f"cic.ext.{target}"
        cmd_mod = importlib.import_module(modname)
        signer_hint = config.get("WALLET_KEY_FILE")
        keys = cmd_mod.list_keys(config, signer_hint)
        if len(keys) > 1:
            print(f"More than one key found, please select one:")
            for idx, key in enumerate(keys):
                print(f"{idx} - {key} ")
            selecting_key = True
            while selecting_key:
                idx = int(input("Select key: "))
                if keys[idx] is not None:
                    key_account_address = keys[idx]
                    selecting_key = False
                else:
                    print("Invalid key, try again")
        else:
            key_account_address = keys[0]

        m = importlib.import_module(f"cic.ext.{target}.start")
        m.extension_start(
            network,
            registry_address=config.get("CIC_REGISTRY_ADDRESS"),
            chain_spec=ChainSpec.from_chain_str(config.get("CHAIN_SPEC")),
            rpc_provider=config.get("RPC_PROVIDER"),
            key_account_address=key_account_address,
        )
    network.load()

    return Contract(
        token=token, proof=proof, meta=meta, attachment=attachment, network=network
    )


def deploy_contract(
    config: Config,
    target: str,
    contract_directory: str,
):

    modname = f"cic.ext.{target}"
    cmd_mod = importlib.import_module(modname)

    writers = init_writers_from_config(config)
    output_directory = os.path.join(contract_directory, "out")
    output_writer_path_meta = output_directory

    metadata_endpoint = config.get("META_URL")

    if metadata_endpoint is not None:
        MetadataRequestsHandler.base_url = metadata_endpoint
        MetadataSigner.gpg_path = "/tmp"
        MetadataSigner.key_file_path = config.get("AUTH_KEYFILE_PATH")
        MetadataSigner.gpg_passphrase = config.get("AUTH_PASSPHRASE")
        writers["proof"] = KeyedWriterFactory(MetadataWriter, HTTPWriter).new
        writers["attachment"] = KeyedWriterFactory(None, HTTPWriter).new
        writers["meta"] = MetadataWriter
        output_writer_path_meta = metadata_endpoint

    ct = Token(path=contract_directory)
    cm = Meta(
        path=contract_directory, writer=writers["meta"](path=output_writer_path_meta)
    )
    ca = Attachment(
        path=contract_directory,
        writer=writers["attachment"](path=output_writer_path_meta),
    )
    cp = Proof(
        path=contract_directory,
        attachments=ca,
        writer=writers["proof"](path=output_writer_path_meta),
    )
    cn = Network(path=contract_directory)

    ca.load()
    ct.load()
    cp.load()
    cm.load()
    cn.load()

    chain_spec = None
    try:
        chain_spec = config.get("CHAIN_SPEC")
        log.debug(f"using CHAIN_SPEC from config: {chain_spec}")
    except KeyError:
        chain_spec = cn.chain_spec
        config.add(chain_spec, "CHAIN_SPEC", exists_ok=True)
        log.debug(f"using CHAIN_SPEC: {str(chain_spec)} from network")

    signer_hint = config.get("WALLET_KEY_FILE")
    (rpc, signer) = cmd_mod.parse_adapter(config, signer_hint)

    target_network_reference = cn.resource(target)
    chain_spec = cn.chain_spec(target)
    log.debug(
        f'found reference {target_network_reference["contents"]} chain spec {chain_spec} for target {target}'
    )
    c = getattr(cmd_mod, "new")(
        chain_spec,
        target_network_reference["contents"],
        cp,
        signer_hint=signer,
        rpc=rpc,
        outputs_writer=writers["ext"](path=output_directory),
    )
    c.apply_token(ct)

    p = ContractProcessor(proof=cp, attachment=ca, metadata=cm, extensions=[c])
    p.process()



# local imports
from cic.contract.components.proof import Proof
from cic.contract.components.meta import Meta
from cic.contract.components.attachment import Attachment
from cic.contract.network import Network
from cic.contract.components.token import Token


def process_args(argparser):
    argparser.add_argument("-f", "--file", type=str, help="add file")
    argparser.add_argument(
        "-d",
        "--directory",
        type=str,
        dest="directory",
        default=".",
        help="cic data directory",
    )


def validate_args(args):
    pass


def execute(config, eargs):
    ct = Token(path=eargs.directory)
    cp = Proof(path=eargs.directory)
    cm = Meta(path=eargs.directory)
    ca = Attachment(path=eargs.directory)
    cn = Network(eargs.directory)

    ct.load()
    cp.load()
    cm.load()
    ca.load()
    cn.load()

    print(
        """[cic.header]
version = {}\n""".format(
            cp.version()
        )
    )
    print("[cic.token]\n{}".format(ct))
    print("[cic.proof]\n{}".format(cp))
    print("[cic.meta]\n{}".format(cm))
    print("[cic.attachment]\n{}".format(ca))
    print("[cic.network]\n{}".format(cn))



# standard imports
import json
import os

# local imports
from cic.contract.base import Data, data_dir
from cic.contract.helpers import select_contract


class Token(Data):
    """Encapsulates the token data used by the extension to deploy and/or register token and token related applications on chain.

    Token details (name, symbol etc) will be used to initialize the token settings when start is called. If load is called instead, any token detail parameters passed to the constructor will be overwritten by data stored in the settings.

    :param path: Settings directory path
    :type path: str
    :param name: Token name
    :type name: str
    :param symbol: Token symbol
    :type symbol: str
    :param precision: Token value precision (number of decimals)
    :type precision: int
    :param supply: Token supply (in smallest precision units)
    :type supply: int
    :param code: Bytecode for token chain application
    :type code: str (hex)
    """

    def __init__(
        self,
        path=".",
        name="Foo Token",
        symbol="FOO",
        precision=6,
        supply=0,
        code=None,
        extra_args=[],
        extra_args_types=[],
        interactive=False,
    ):
        super(Token, self).__init__()
        self.name = name
        self.symbol = symbol
        self.supply = supply
        self.precision = precision
        self.code = code
        self.extra_args = extra_args
        self.extra_args_types = extra_args_types
        self.path = path
        self.token_path = os.path.join(self.path, "token.json")

        if interactive:
            contract = select_contract()
            self.code = contract["bin_path"]
            self.extra_args = contract["extra_args"]
            self.extra_args_types = contract["extra_args_types"]

            self.name = input(f"Enter Token Name ({self.name}): ") or self.name
            self.symbol = input(f"Enter Token Symbol ({self.symbol}): ") or self.symbol
            self.precision = input(f"Enter Token Precision ({self.precision}): ") or self.precision
            self.supply = input(f"Enter Token Supply ({self.supply}): ") or self.supply

    def load(self):
        """Load token data from settings."""
        super(Token, self).load()

        with open(self.token_path, "r", encoding="utf-8") as f:
            o = json.load(f)

        self.name = o["name"]
        self.symbol = o["symbol"]
        self.precision = o["precision"]
        self.code = o["code"]
        self.supply = o["supply"]
        extras = []
        extra_types = []
        token_extras: list = o["extra"]
        if token_extras:
            for idx, token_extra in enumerate(token_extras):
                arg = token_extra.get("arg")
                arg_type = token_extra.get("arg_type")
                if arg and arg_type:
                    extras.append(arg)
                    extra_types.append(arg_type)
                elif (arg and not arg_type) or (not arg and arg_type):
                    raise ValueError(
                        f"Extra contract args must have a 'arg' and 'arg_type', Please check {self.token_path}:extra[{idx}] "
                    )
        self.extra_args = extras
        self.extra_args_types = extra_types
        self.inited = True

    def start(self):
        """Initialize token settings from arguments passed to the constructor and/or template."""
        super(Token, self).load()

        token_template_file_path = os.path.join(
            data_dir, f"token_template_v{self.version()}.json"
        )
        with open(token_template_file_path, encoding="utf-8") as f:
            o = json.load(f)
        o["name"] = self.name
        o["symbol"] = self.symbol
        o["precision"] = self.precision
        o["code"] = self.code
        o["supply"] = self.supply
        extra = []
        for idx, extra_arg in enumerate(self.extra_args):
            extra.append({"arg": extra_arg, "arg_type": self.extra_args_types[idx]})
        if len(extra) != 0:
            o["extra"] = extra

        with open(self.token_path, "w", encoding="utf-8") as f:
            json.dump(o, f, sort_keys=True, indent="\t")

    def __str__(self):
        s = f"name = {self.name}\n"
        s += f"symbol = {self.symbol}\n"
        s += f"precision = {self.precision}\n"
        s += f"supply = {self.supply}\n"
        for idx, extra in enumerate(self.extra_args):
            s += f"extra_args[{idx}]({self.extra_args_types[idx]}) = {extra}\n"
        return s



# standard imports
import copy
import json
import logging

# external imports
from chainlib.chain import ChainSpec
from chainlib.eth.address import is_address, to_checksum_address
from chainlib.eth.connection import RPCConnection
from chainlib.eth.contract import ABIContractEncoder, ABIContractType
from chainlib.eth.gas import OverrideGasOracle
from chainlib.eth.nonce import RPCNonceOracle
from chainlib.eth.tx import Tx, TxFactory, TxFormat, receipt
from eth_address_declarator import Declarator
from eth_address_declarator.declarator import AddressDeclarator
from eth_token_index import TokenUniqueSymbolIndex
from giftable_erc20_token import GiftableToken
from hexathon import add_0x, strip_0x

# local imports
from cic.ext.eth.rpc import parse_adapter, list_keys
from cic.extension import Extension


logg = logging.getLogger(__name__)


class CICEth(Extension):
    def __init__(
        self,
        chain_spec,
        resources,
        proof,
        signer=None,
        rpc=None,
        outputs_writer=None,
        fee_oracle=None,
    ):

        """Implementation for the eth extension.

        The state of the resources instance will be modified.

        :param chain_spec: Chain Spec that extension will operate for
        :type chain_spec: chainlib.chain.ChainSpec
        :param resources: Chain application resources to deploy or interface with
        :type resources: dict
        :param proof: Proof object to publish
        :type proof: cic.proof.Proof
        :param signer: Signer capable of generating signatures for chain aplication deployments
        :type signer: funga.signer.Signer
        :param rpc: RPC adapter capable of submitting and querying the chain network node
        :type rpc: chainlib.connection.RPCConnection
        :param outputs_writer: Writer interface receiving the output of the processor
        :type outputs_writer: cic.writers.OutputWriter
        :param fee_oracle: Fee oracle required by signer
        :type fee_oracle: chainlib.fee.FeeOracle
        """
        super(CICEth, self).__init__(
            chain_spec,
            resources,
            proof,
            signer=signer,
            rpc=rpc,
            outputs_writer=outputs_writer,
        )
        self.fee_oracle = fee_oracle
        self.tx_format = TxFormat.RAW_ARGS
        if self.rpc is not None:
            self.tx_format = TxFormat.JSONRPC
        elif self.signer is not None:
            self.tx_format = TxFormat.RLP_SIGNED

    def __detect_arg_type(self, v):
        typ = None
        try:
            int(v, 10)
            typ = ABIContractType.UINT256
        except TypeError:
            pass

        if typ is None:
            try:
                vv = strip_0x(v)
                if is_address(vv):
                    typ = ABIContractType.ADDRESS
                else:
                    typ = ABIContractType.BYTES32
            except ValueError:
                pass

        if typ is None:
            try:
                v.encode("utf-8")
                typ = ABIContractType.STRING
            except ValueError:
                pass

        if typ is None:
            raise ValueError(
                f"cannot automatically determine type for value {v}"
            )

        logg.info(f"argument {v} parsed as abi contract type {typ.value}")

        return typ

    def __order_args(self):
        args = [
            self.token_details["name"],
            self.token_details["symbol"],
            self.token_details["precision"],
        ]
        args_types = [
            ABIContractType.STRING.value,
            ABIContractType.STRING.value,
            ABIContractType.UINT256.value,
        ]

        for i, x in enumerate(self.token_details["extra"]):
            args.append(x)
            typ = None
            if self.token_details["extra_types"] is not None:
                typ = self.token_details["extra_types"][i]
            else:
                typ = self.__detect_arg_type(x)
            args_types.append(typ)

        positions = self.token_details["positions"]
        if positions is None:
            positions = list(range(len(args)))

        return (args, args_types, positions)

    def add_outputs(self, k, v):
        """Adds given key/value pair to outputs array.

        :param k: Output key
        :type k: str
        :param v: Output value
        :param v: bytes or str
        """
        logg.debug(f"adding outputs {k} {v}")
        self.outputs.append((k, v))

    def get_outputs(self):
        """Get wrapper for outputs captured from processing.

        :rtype: list of tuples
        :return: Captured outputs
        """
        return self.outputs

    def process_token(self, writer=None):
        """Deploy token, and optionally mint token supply to token deployer account.

        :param writer: Writer interface receiving the output of the processor step
        :type writer: cic.writers.OutputWriter
        """
        if writer is None:
            writer = self.outputs_writer

        (args, args_types, positions) = self.__order_args()

        enc = ABIContractEncoder()

        for i in positions:
            getattr(enc, args_types[i])(args[i])

        code = enc.get()
        if self.token_code is not None:
            code = self.token_code + code

        logg.debug(f"resource {self.resources}")
        signer_address = add_0x(
            to_checksum_address(self.resources["token"]["key_account"])
        )
        nonce_oracle = None
        if self.rpc is not None:
            nonce_oracle = RPCNonceOracle(signer_address, conn=self.rpc)

        c = TxFactory(
            self.chain_spec,
            signer=self.signer,
            nonce_oracle=nonce_oracle,
            gas_oracle=self.fee_oracle,
        )
        tx = c.template(signer_address, None, use_nonce=True)
        tx = c.set_code(tx, code)
        o = c.finalize(tx, self.tx_format)

        token_address_tx = None
        r = None
        if self.rpc is not None:
            r = self.rpc.do(o[1])
            token_address_tx = r
            o = self.rpc.wait(r)
            o = Tx.src_normalize(o)
            self.token_address = o["contract_address"]
        elif self.signer is not None:
            r = o[1]
            token_address_tx = r

        if r is None:
            r = code
        writer.write("token", r.encode("utf-8"))
        writer.write("token_address", self.token_address.encode("utf-8"))
        self.add_outputs("token", r)

        if int(self.token_details["supply"]) > 0:
            c = GiftableToken(
                self.chain_spec,
                signer=self.signer,
                nonce_oracle=nonce_oracle,
                gas_oracle=self.fee_oracle,
            )
            o = c.mint_to(
                self.token_address,
                self.resources["token"]["key_account"],
                self.resources["token"]["key_account"],
                self.token_details["supply"],
            )
            r = None
            if self.rpc is not None:
                r = self.rpc.do(o[1])
                self.rpc.wait(r)
                writer.write("token_supply", r.encode("utf-8"))
            elif self.signer is not None:
                r = o[1]
                writer.write(
                    "token_supply", json.dumps(r, separators=(",", ":")).encode("utf-8")
                )
            else:
                r = o
                writer.write("token_supply", r.encode("utf-8"))

        return token_address_tx

    def process_token_index(self, writer=None):
        """Register deployed token with token index.

        :param writer: Writer interface receiving the output of the processor step
        :type writer: cic.writers.OutputWriter
        """
        if writer is None:
            writer = self.outputs_writer

        signer_address = add_0x(
            to_checksum_address(self.resources["token_index"]["key_account"])
        )
        contract_address = add_0x(
            to_checksum_address(self.resources["token_index"]["reference"])
        )

        gas_oracle = OverrideGasOracle(
            limit=TokenUniqueSymbolIndex.gas(), conn=self.rpc
        )
        nonce_oracle = None
        if self.rpc is not None:
            nonce_oracle = RPCNonceOracle(add_0x(signer_address), conn=self.rpc)
        c = TokenUniqueSymbolIndex(
            self.chain_spec,
            signer=self.signer,
            nonce_oracle=nonce_oracle,
            gas_oracle=gas_oracle,
        )

        o = c.register(
            contract_address,
            signer_address,
            self.token_address,
            tx_format=self.tx_format,
        )
        r = None
        if self.rpc is not None:
            r = self.rpc.do(o[1])
            self.rpc.wait(r)
        elif self.signer is not None:
            r = o[1]
        else:
            r = o

        writer.write("token_index", r.encode("utf-8"))
        self.add_outputs("token_index", r)
        return r

    def process_address_declarator(self, writer=None):
        """Register token proofs with address declarator.

        :param writer: Writer interface receiving the output of the processor step
        :type writer: cic.writers.OutputWriter
        """
        if writer is None:
            writer = self.outputs_writer

        signer_address = add_0x(
            to_checksum_address(self.resources["address_declarator"]["key_account"])
        )
        contract_address = add_0x(
            to_checksum_address(self.resources["address_declarator"]["reference"])
        )

        gas_oracle = OverrideGasOracle(limit=AddressDeclarator.gas(), conn=self.rpc)
        nonce_oracle = None
        if self.rpc is not None:
            nonce_oracle = RPCNonceOracle(signer_address, conn=self.rpc)
        c = Declarator(
            self.chain_spec,
            signer=self.signer,
            nonce_oracle=nonce_oracle,
            gas_oracle=gas_oracle,
        )

        results = []
        # (main_proof, all_proofs) = self.proof.get()

        # for proof in all_proofs:
        # logg.debug('proof {} '.format(proof))

        (k, v) = self.proof.root()

        fk = "address_declarator_" + k
        o = c.add_declaration(
            contract_address,
            signer_address,
            self.token_address,
            k,
            tx_format=self.tx_format,
        )
        r = None
        if self.rpc is not None:
            r = self.rpc.do(o[1])
            self.rpc.wait(r)
        elif self.signer is not None:
            r = o[1]
        else:
            r = o
        self.add_outputs(fk, r)
        results.append(r)
        v = r.encode("utf-8")
        if writer is not None:
            writer.write(fk, v)

        return results

    def prepare_extension(self):
        """Sets token address for extension if defined in settings."""
        super(CICEth, self).prepare_extension()

        if self.token_address is not None:
            self.token_address = add_0x(to_checksum_address(self.token_address))


def new(chain_spec, resources, proof, signer_hint=None, rpc=None, outputs_writer=None):
    """Convenience function to enable object instantiation through predictable module symbol

    See CICEth constructor for details.
    """
    return CICEth(
        chain_spec,
        resources,
        proof,
        signer=signer_hint,
        rpc=rpc,
        outputs_writer=outputs_writer,
    )



from __future__ import annotations

# standard import
import logging

from chainlib.cli.config import Config

# local imports
from cic.contract.contract import deploy_contract, generate_contract, load_contract

log = logging.getLogger(__name__)


def process_args(argparser):
    argparser.add_argument(
        "--skip-gen", action="store_true", default=False, help="Skip Generation"
    )
    argparser.add_argument(
        "--skip-deploy",
        action="store_true",
        help="Skip Deployment",
    )
    argparser.add_argument(
        "--target",
        default="eth",
        help="Contract Target (eth)",
    )
    argparser.add_argument(
        "path",
        type=str,
        help="Path to generate/use contract deployment info",
    )
    argparser.add_argument(
        "-p",
        type=str,
        help="RPC Provider (http://localhost:8545)",
    )
    argparser.add_argument(
        "-y",
        type=str,
        help="Wallet Keystore",
    )


def validate_args(_args):
    pass


ExtraArgs = {"skip_gen": str, "skip_deploy": str, "target": str, "path": str, "p": str}


def execute(config: Config, eargs: ExtraArgs):
    directory = eargs.path
    target = eargs.target
    skip_gen = eargs.skip_gen
    skip_deploy = eargs.skip_deploy
    wallet_keystore = eargs.y
    if wallet_keystore:
        config.add(wallet_keystore, "WALLET_KEY_FILE", exists_ok=True)

    if not skip_gen:
        contract = generate_contract(directory, [target], config, interactive=True)
    else:
        contract = load_contract(directory)

    print(contract)

    print(f"Meta: {config.get('META_URL')}")
    print(f"ChainSpec: {config.get('CHAIN_SPEC', contract.network.chain_spec(target))}")
    print(f"RPC: {config.get('RPC_PROVIDER')}\n")

    if not skip_deploy:
        ready_to_deploy = input("Are you ready to Deploy? (y/n): ")
        if ready_to_deploy == "y":
            deploy_contract(
                config=config,
                contract_directory=directory,
                target=target,
            )
            print("Deployed")
        else:
            print("Skipping deployment")


if __name__ == "__main__":
    # execute()
    print("Not Implemented")



# standard imports
import logging

# external imports
from hexathon import valid as valid_hex

# local imports
from cic.writers import StdoutWriter
from cic.contract.components.token import Token

logg = logging.getLogger(__name__)


class Extension:
    """Base class adapter to initialize, serialize and publish extension-specific token resources.

    :param chain_spec: Chain Spec that extension will operate for
    :type chain_spec: chainlib.chain.ChainSpec
    :param resources: Chain application resources to deploy or interface with
    :type resources: dict
    :param proof: Proof object to publish
    :type proof: cic.proof.Proof
    :param signer: Signer capable of generating signatures for chain aplication deployments
    :type signer: funga.signer.Signer
    :param rpc: RPC adapter capable of submitting and querying the chain network node
    :type rpc: chainlib.connection.RPCConnection
    :param writer: Writer interface receiving the output of the processor
    :type writer: cic.writers.OutputWriter
    """

    def __init__(
        self,
        chain_spec,
        resources,
        proof,
        signer=None,
        rpc=None,
        outputs_writer=StdoutWriter(),
    ):
        self.resources = resources
        self.proof = proof
        self.chain_spec = chain_spec
        self.signer = signer
        self.rpc = rpc
        self.token_details = None
        self.token_address = None
        self.token_code = None
        self.outputs = []
        self.outputs_writer = outputs_writer

    # TODO: apply / prepare token can be factored out
    def apply_token(self, token: Token):
        """Initialize extension with token data from settings.

        :param token: Token object
        :type token: cic.token.Token
        :rtype: dict
        :returns: Token data state of extension after load
        """
        return self.prepare_token(
            token.name,
            token.symbol,
            token.precision,
            token.code,
            token.supply,
            token.extra_args,
            token.extra_args_types,
        )

    def prepare_token(
        self,
        name,
        symbol,
        precision,
        code,
        supply,
        extra=None,
        extra_types=None,
        positions=None,
    ):
        """Initialize extension token data.

        :param name: Token name
        :type name: str
        :param symbol: Token symbol
        :type symbol: str
        :param precision: Token value precision (number of decimals)
        :type precision: int
        :param code: Bytecode for token chain application
        :type code: str (hex)
        :param supply: Token supply (in smallest precision units)
        :type supply: int
        :param extra: Extra parameters to pass to token application constructor
        :type extra: list
        :param extra_types: Type specifications for extra parameters
        :type extra_types: list
        :param positions: Sequence of parameter indices to pass to application constructor
        :type positions: list
        :rtype: dict
        :returns: Token data state of extension after load
        """
        self.token_details = {
            "name": name,
            "symbol": symbol,
            "precision": precision,
            "code": code,
            "supply": supply,
            "extra": extra or [],
            "extra_types": extra_types or [],
            "positions": positions,
        }
        logg.debug(f"token details: {self.token_details}")
        return self.token_details

    def prepare_extension(self):
        """Prepare extension for publishing (noop)"""


    def parse_code_as_file(self, v):
        """Helper method to load application bytecode from file into extensions token data state.

        Client code should call load_code instead.

        :param v: File path
        :type v: str
        """
        try:
            f = open(v, "r", encoding="utf-8")
            r = f.read()
            f.close()
            self.parse_code_as_hex(r)
        except FileNotFoundError as e:
            logg.debug(f"could not parse code as file: {e}")
        except IsADirectoryError as e:
            logg.debug(f"could not parse code as file: {e}")

    def parse_code_as_hex(self, v):
        """Helper method to load application bytecode from hex data into extension token data state.

        Client code should call load_code instead.

        :param v: Bytecode as hex
        :type v: str
        """
        try:
            self.token_code = valid_hex(v)
        except ValueError as e:
            logg.debug(f"could not parse code as hex: {e}")

    def load_code(self, hint=None):
        """Attempt to load token application bytecode using token settings.

        :param hint: If "hex", will interpret code in settings as literal bytecode
        :type hint: str
        :rtype: str (hex)
        :return: Bytecode loaded into extension token data state
        """
        code = self.token_details["code"]
        if hint == "hex":
            self.token_code = valid_hex(code)

        for m in [
            self.parse_code_as_hex,
            self.parse_code_as_file,
        ]:
            m(code)
            if self.token_code is not None:
                break

        if self.token_code is None:
            raise RuntimeError("could not successfully parse token code")

        return self.token_code

    def process(self, writer=None):
        """Adapter used by Processor to process the extensions implementing the Extension base class.

        Requires either token address or a valid token code reference to have been included in settings.
        If token address is not set, the token application code will be deployed.

        :param writer: Writer to use for publishing.
        :type writer: cic.writers.OutputWriter
        :rtype: tuple
        :return: Token address, token symbol
        """
        if writer is None:
            writer = self.outputs_writer

        tasks = []
        self.token_address = self.resources["token"]["reference"]

        # TODO: get token details when token address is not none
        if self.token_address is None:
            if self.token_details["code"] is None:
                raise RuntimeError("neither token address nor token code has been set")
            self.load_code()
            tasks.append("token")

        for k in self.resources.keys():
            if k == "token":
                continue
            if self.resources[k]["reference"] is not None:
                tasks.append(k)

        self.prepare_extension()

        for task in tasks:
            logg.debug(f"extension adapter process {task}")
            _r = getattr(self, "process_" + task)(writer=writer)

        return (self.token_address, self.token_details.get("symbol"))



import os
import shutil

default_module_configs = os.path.join(os.path.dirname(os.path.realpath(__file__)), '..', 'config')

def ensure_base_configs(config_dir: str):
    """
    Ensure that the base configs are present.
    """
    if not os.path.exists(config_dir):
        os.makedirs(config_dir)
    for f in os.listdir(default_module_configs):
        if not os.path.exists(os.path.join(config_dir, f)):
            shutil.copytree(os.path.join(default_module_configs, f), os.path.join(config_dir, f))
    



# standard imports
import json
import logging
import os
import tempfile

# external imports
from hexathon import strip_0x
from cic_types import MetadataPointer
from cic_types.processor import generate_metadata_pointer

# local imports
from cic.contract.base import Data, data_dir
from cic.utils import object_to_str


logg = logging.getLogger(__name__)


class Proof(Data):
    """Proof handles the immutable token proof data mapped to the initial token deployment.

    It processes inputs from the proof.json file in the session directory.

    Optionally, attachment objects can be added to the proof. If added, the resulting proof digest will consists of the attachment digests added to the root digest. These are then are deterministically ordered, regardless of which order attachments were given to the constructor.

    :param path: Path to settings directory
    :type path: str
    :param attachments: List of attachment objects to include in the proof
    :type attachments: cic.attachment.Attachment
    :param writer: Writer interface receiving the output of the processor
    :type writer: cic.writers.OutputWriter
    """

    def __init__(
        self,
        path=".",
        description="",
        namespace="ge",
        issuer="",
        attachments=None,
        writer=None,
        interactive=False,
    ):
        super(Proof, self).__init__()
        self.proofs = []
        self.namespace = namespace
        self.description = description
        self.issuer = issuer
        self.path = path
        self.writer = writer
        self.extra_attachments = attachments
        self.attachments = {}
        self.proof_path = os.path.join(self.path, "proof.json")
        self.temp_proof_path = tempfile.mkstemp()[1]

        if interactive:
            self.description = (
                input(f"Enter Proof Description ({self.description}): ") or self.description
            )
            self.namespace = (
                input(f"Enter Proof Namespace ({self.namespace}): ") or self.namespace
            )
            self.issuer = input(f"Enter Proof Issuer ({self.issuer}): ") or self.issuer

    def load(self):
        """Load proof data from settings."""
        super(Proof, self).load()

        f = open(self.proof_path, "r")
        o = json.load(f)
        f.close()

        self.set_version(o["version"])
        self.description = o["description"]
        self.namespace = o["namespace"]
        self.issuer = o["issuer"]
        self.proofs = o["proofs"]

        if self.extra_attachments != None:
            a = self.extra_attachments.asdict()
            for k in a.keys():
                self.attachments[k] = a[k]

        hshs = self.__get_ordered_hashes()
        self.proofs = list(map(strip_0x, hshs))

        self.inited = True

    def start(self):
        """Initialize proof settings from template."""
        super(Proof, self).start()

        proof_template_file_path = os.path.join(
            data_dir, f"proof_template_v{self.version()}.json"
        )

        with open(proof_template_file_path, "r", encoding="utf-8") as f:
            o = json.load(f)

        o["issuer"] = self.issuer
        o["description"] = self.description
        o["namespace"] = self.namespace

        with open(self.proof_path, "w", encoding="utf-8") as f:
            json.dump(o, f, sort_keys=True, indent="\t")

    def asdict(self):
        """Output proof state to dict."""
        return {
            "version": self.version(),
            "namespace": self.namespace,
            "description": self.description,
            "issuer": self.issuer,
            "proofs": self.proofs,
        }

    # TODO: the efficiency of this method could probably be improved.
    def __get_ordered_hashes(self):
        ks = list(self.attachments.keys())
        ks.sort()

        return ks

    #    def get(self):
    #        hsh = self.hash(b).hex()
    #        self.attachments[hsh] = self.temp_proof_path
    #        logg.debug('cbor of {} is {} hashes to {}'.format(v, b.hex(), hsh))

    def root(self):
        """Calculate the root digest from the serialized proof object."""
        v = self.asdict()
        # b = cbor2.dumps(v)
        b = json.dumps(v, separators=(",", ":"))

        with open(self.temp_proof_path, "w", encoding="utf-8") as f:
            f.write(b)

        b = b.encode("utf-8")
        k = self.hash(b)

        return (k.hex(), b)

    def process(self, token_address=None, token_symbol=None, writer=None):
        """Serialize and publish proof.

        See cic.processor.Processor.process
        """
        if writer is None:
            writer = self.writer

        (k, v) = self.root()
        writer.write(k, v)
        root_key = k

        token_symbol_bytes = token_symbol.encode("utf-8")
        k = generate_metadata_pointer(
            token_symbol_bytes, MetadataPointer.TOKEN_PROOF_SYMBOL
        )
        writer.write(k, v)

        token_address_bytes = bytes.fromhex(strip_0x(token_address))
        k = generate_metadata_pointer(token_address_bytes, MetadataPointer.TOKEN_PROOF)
        writer.write(k, v)

        #        (hsh, hshs) = self.get()
        # hshs = list(map(strip_0x, hshs))
        #        hshs_bin = list(map(bytes.fromhex, hshs))
        #        hshs_cat = b''.join(hshs_bin)

        #        f = open(self.temp_proof_path, 'rb')
        #        v = f.read()
        #        f.close()
        #        writer.write(hsh, v)

        #        r = self.hash(hshs_cat)
        #        r_hex = r.hex()

        # logg.debug('generated proof {} for hashes {}'.format(r_hex, hshs))

        # writer.write(r_hex, hshs_cat)

        o = self.asdict()
        with open(self.proof_path, "w", encoding="utf-8") as f:
            json.dump(o, f, sort_keys=True, indent="\t")

        return root_key

    def __str__(self):
        return object_to_str(
            self, ["description", "issuer", "namespace", "version()", "proofs"]
        )



# standard imports
from getpass import getpass
import logging
import os
import stat

# external imports
from funga.eth.keystore.dict import DictKeystore
from funga.eth.signer import EIP155Signer
from chainlib.cli import Wallet
from chainlib.eth.cli import Rpc

# local imports
from cic.keystore import KeystoreDirectory


logg = logging.getLogger(__name__)


class EthKeystoreDirectory(DictKeystore, KeystoreDirectory):
    """Combination of DictKeystore and KeystoreDirectory

    TODO: Move to funga
    """


def get_passphrase():
    return getpass('Enter passphrase: ')

def parse_adapter(config, signer_hint):
    """Determine and instantiate signer and rpc from configuration.

    If either could not be determined, None is returned.

    :param config: Configuration object implementing the get() method
    :type config: dict, object with get()
    :param signer_hint: Signer resource description (e.g. keystore directory)
    :type signer_hint: str
    :rtype: tuple; chainlib.connection.RPCConnection, funga.signer.Signer
    :return: RPC interface, signer interface
    """
    keystore = None
    if signer_hint is None:
        logg.info("signer hint missing")
        return None
    st = os.stat(signer_hint)
    if stat.S_ISDIR(st.st_mode):
        logg.debug("signer hint is directory")
        keystore = EthKeystoreDirectory()
        keystore.process_dir(signer_hint, password_retriever=get_passphrase)

    w = Wallet(EIP155Signer, keystore=keystore)
    signer = EIP155Signer(keystore)
    rpc = Rpc(wallet=w)
    rpc.connect_by_config(config)

    return (rpc.conn, signer)

# TODO Find a better place for this
def list_keys(config, signer_hint):
    keystore = None
    if signer_hint is None:
        logg.info("signer hint missing")
        return None
    st = os.stat(signer_hint)
    if stat.S_ISDIR(st.st_mode):
        logg.debug("signer hint is directory")
        keystore = EthKeystoreDirectory()
        keystore.process_dir(signer_hint, default_password=config.get('WALLET_PASSPHRASE', ''), password_retriever=get_passphrase)
        
    return keystore.list()



# standard imports
import os
import json
import logging

# external imports
from funga.error import (
        DecryptError,
        KeyfileError,
        )
from funga.keystore import Keystore

logg = logging.getLogger(__name__)


class KeystoreDirectory(Keystore):

    def process_dir(self, path, password_retriever=None, default_password=''):
        for v in os.listdir(path):
            fp = os.path.join(path, v)
            try:
                self.import_keystore_file(fp, password=default_password)
            except IsADirectoryError:
                pass
            except KeyfileError as e:
                logg.warning(f'file {fp} could not be parsed as keyfile: {e}')
            except DecryptError as e:
                if password_retriever is None:
                    raise e
                password = password_retriever()
                self.import_keystore_file(fp, password=password)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cic',
 'cic.cmd',
 'cic.contract',
 'cic.contract.components',
 'cic.ext.eth',
 'cic.runnable',
 'cmd',
 'eth']

package_data = \
{'': ['*'], 'cic': ['data/*', 'data/config/*', 'schema/*']}

modules = \
['cic_cmd']
install_requires = \
['cbor2==5.4.1',
 'chainlib>=0.0.17,<0.1.0',
 'cic-types>=0.2.1a8,<0.3.0',
 'confini>=0.5.3,<0.6.0',
 'funga-eth>=0.5.5,<0.6.0']

extras_require = \
{'eth': ['chainlib-eth>=0.0.25,<0.1.0',
         'eth-token-index>=0.2.4,<0.3.0',
         'eth-address-index>=0.2.4,<0.3.0',
         'okota>=0.2.5,<0.3.0',
         'cic_eth_registry>=0.6.6,<0.7.0',
         'cic_contracts>=0.0.5,<0.1.0']}

entry_points = \
{'console_scripts': ['cic = cic.runnable.cic_cmd:main']}

setup_kwargs = {
    'name': 'cic-cli',
    'version': '0.2.0',
    'description': 'Generic cli tooling for the CIC token network',
    'long_description': '# CIC Token Deployment Tool\n[![Status](https://ci.grassecon.net/api/badges/grassrootseconomics/cic/status.svg?ref=refs/heads/master)](https://ci.grassecon.net/grassrootseconomics/cic)\n[![Version](https://img.shields.io/pypi/v/cic?color=green)](https://pypi.org/project/cic/)\n\nCIC-CLI provides tooling to generate and publish metadata in relation to\ntoken deployments.\n\n```shell\npip install cic-cli[eth]\n```\n## Usage\n### Using the wizard  \n```\n# Local\ncic wizard ./somewhere -c ./config/docker\n\n# Test Net\ncic wizard ./somewhere -c ./config/testnet\n\n```\n### Modular\nSome of the concepts described below assume familiarity with base\nconcepts of the CIC architecture. Please refer to the appropriate\ndocumentation for more information.\n\nTo initialize a new token deployment for the EVM:\n\n```shell\ncic init --target eth --name <token_name> --symbol <token_symbol> --precision <token_value_precision> <settings_folder>\n```\n\nTo automatically fill in settings detected in the network for the EVM:\n\n```shell\ncic ext --registry <contract_registry_address> -d <settings_folder> -i <chain_spec> -p <rpc_endpoint> eth\n```\n\n\n## Structure of the components\n\n![image](./doc/sphinx/components.svg)\n\nCIC-CLI is designed to interface any network type backend. The current\nstate of the package contains interface to EVM only. Thus, the examples\nbelow are limited to the context of the EVM.\n\n## Development\n### Requirements\n - Install [poetry](https://python-poetry.org/docs/#installation) \n\n### Setup\n\n```\n poetry install -E eth\n```\n\n### Running the CLI\n\n```bash\n poetry run cic -h\n```\n\n```bash\n poetry run cic wizard ./somewhere -c ./config/docker\n```\n### Importing a wallet from metamask\n- Export the accounts private key [Instructions](https://metamask.zendesk.com/hc/en-us/articles/360015289632-How-to-Export-an-Account-Private-Key)\n- Save the private key to a file\n- Run `eth-keyfile -k <file> > ~/.config/cic/keystore/keyfile.json`\n\n###  Port Forwarding\n<details>\n<summary>Install Kubectl</summary>\n\n```bash\nsudo apt-get update\nsudo apt-get install -y apt-transport-https ca-certificates curl\nsudo curl -fsSLo /usr/share/keyrings/kubernetes-archive-keyring.gpg https://packages.cloud.google.com/apt/doc/apt-key.gpg\necho "deb [signed-by=/usr/share/keyrings/kubernetes-archive-keyring.gpg] https://apt.kubernetes.io/ kubernetes-xenial main" | sudo tee /etc/apt/sources.list.d/kubernetes.list\nsudo apt-get update\nsudo apt-get install -y kubectl\n```\n</details>\n\n- Download testnet cluster config from https://cloud.digitalocean.com/kubernetes/clusters\n- Move the config to `$HOME/.kube/`\n- Run `kubectl -n grassroots --kubeconfig=$HOME/.kube/<config_file_name>.yaml get pods`  \n- Copy the name of the meta pod (e.g `cic-meta-server-67dc7c6468-8rhdq`)\n- Port foward the meta pod to the local machine using `kubectl port-forward pods/<name_of_meta_pod> 6700:8000 -n grassroots --kubeconfig=$HOME/.kube/<config_file_name>.yaml`\n- Clone this repository to your local machine\n- Run `poetry install -E eth` in the repo root\n- Open `./config/testnet/config.ini` and change\n  - [auth]keyfile_path \n  - [wallet]key_file\n- Open a new terminal and run `poetry run cic wizard -c ./config/testnet ./somewhere` \n### Tests\n\n```\npoetry run pytest\n```\n',
    'author': 'Louis Holbrook',
    'author_email': 'dev@holbrook.no',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://git.grassecon.net/cicnet/cic-cli',
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# standard imports
import logging

logg = logging.getLogger(__name__)


class ContractProcessor:
    """Drives the serialization and publishing of contracts, proofs and metadata for the token.

    :param proof: Proof object to publish
    :type proof: cic.proof.Proof
    :param attachment: Attachment object to publish
    :type attachment: cic.attachment.Attachment
    :param metadata: Metadata object to publish
    :type metadata: cic.meta.Meta
    :param writer: Writer interface receiving the output of the processor
    :type writer: cic.writers.OutputWriter
    :param extensions: Extension contexts to publish to
    :type extensions: list of cic.extension.Extension
    """

    def __init__(
        self,
        proof=None,
        attachment=None,
        metadata=None,
        outputs_writer=None,
        extensions=[],
    ):
        self.token_address = None
        self.extensions = extensions
        self.cores = {
            "metadata": metadata,
            "attachment": attachment,
            "proof": proof,
        }
        self.outputs = []
        self.__outputs_writer = outputs_writer

    def writer(self):
        """Return the writer instance that the process is using.

        :rtype: cic.writers.OutputWriter
        :return: Writer
        """
        return self.__outputs_writer

    def get_outputs(self):
        """Return all written outputs.

        This will return nothing unless the process method has been executed.

        :rtype: bytes
        :return: Outputs
        """
        outputs = []
        for ext in self.extensions:
            outputs += ext.outputs
        outputs += self.outputs
        return outputs

    def process(self, writer=None):
        """Serializes and publishes all token data.

        Calls the process method on each extension. For each extension, the process method on attachment, proof and metadata, in that order, for any of them that have provided at processor object instantiation.

        All output written to the publish writer will also be cached so that it subsequently be recalled using the get_outputs method.

        :param writer: Writer to use for publishing.
        :type writer: cic.writers.OutputWriter
        """

        tasks = [
            "attachment",
            "proof",
            "metadata",
        ]

        for ext in self.extensions:
            (token_address, token_symbol) = ext.process()

            for task in tasks:
                a = self.cores.get(task)
                if a is None:
                    logg.debug(f'skipping missing task receiver "{task}"')
                    continue
                v = a.process(
                    token_address=token_address,
                    token_symbol=token_symbol,
                    writer=self.__outputs_writer,
                )
                self.outputs.append(v)



# external imports
from chainlib.eth.connection import RPCConnection
from cic_eth_registry import CICRegistry


def extension_start(network, *args, **kwargs):
    """Called by the "export" cli tool subcommand for initialization of the eth extension.

    :param network: Network object to read and write settings from
    :type network: cic.network.Network
    """
    CICRegistry.address = kwargs.get("registry_address")
    key_account_address = kwargs.get("key_account_address")
    RPCConnection.register_location(
        kwargs.get("rpc_provider"), kwargs.get("chain_spec")
    )
    conn = RPCConnection.connect(kwargs.get("chain_spec"))

    registry = CICRegistry(kwargs.get("chain_spec"), conn)

    address_declarator = registry.by_name("AddressDeclarator")
    network.resource_set(
        "eth", "address_declarator", address_declarator, key_account=key_account_address
    )

    token_index = registry.by_name("TokenRegistry")
    network.resource_set(
        "eth", "token_index", token_index, key_account=key_account_address
    )

    network.resource_set("eth", "token", None, key_account=key_account_address)

    network.set("eth", kwargs["chain_spec"])
    network.save()



# standard imports
import base64
import logging
import os
import sys
import urllib.request
import json

from cic_types.ext.metadata import MetadataRequestsHandler, MetadataPointer

logg = logging.getLogger(__name__)


class OutputWriter:
    def __init__(self, *args, **kwargs):
        pass

    def write(self, k, v):
        raise NotImplementedError()


class StdoutWriter(OutputWriter):
    def write(self, k, v):
        sys.stdout.write(f"{k}\t{v}\n")


class KVWriter(OutputWriter):
    def __init__(self, path=None, *args, **kwargs):
        try:
            os.stat(path)
        except FileNotFoundError:
            os.makedirs(path)
        self.path = path

    def write(self, k, v):
        fp = os.path.join(self.path, str(k))
        logg.debug(f"path write {fp} {str(v)}")
        f = open(fp, "wb")
        f.write(v)
        f.close()


class HTTPWriter(OutputWriter):
    def __init__(self, path=None, *args, **kwargs):
        super(HTTPWriter, self).__init__(*args, **kwargs)
        self.path = path

    def write(self, k, v):
        path = self.path
        if k is not None:
            path = os.path.join(path, k)
        logg.debug(f"http writer post {path} \n key: {k}, value: {v}")
        rq = urllib.request.Request(path, method="POST", data=v)
        r = urllib.request.urlopen(rq)
        logg.info(f"http writer submitted at {r.read()}")


class KeyedWriter(OutputWriter):
    def __init__(self, writer_keyed, writer_immutable):
        self.writer_keyed = writer_keyed
        self.writer_immutable = writer_immutable

    def write(self, k, v):
        logg.debug(f"writing keywriter key: {k} value: {v}")
        if isinstance(v, str):
            v = v.encode("utf-8")
        if self.writer_keyed is not None:
            self.writer_keyed.write(k, v)
        if self.writer_immutable is not None:
            self.writer_immutable.write(None, v)


class KeyedWriterFactory:
    def __init__(
        self, key_writer_constructor, immutable_writer_constructor, *args, **kwargs
    ):
        self.key_writer_constructor = key_writer_constructor
        self.immutable_writer_constructor = immutable_writer_constructor
        self.x = {}
        for k, v in kwargs.items():
            logg.debug(f"adding key {k} t keyed writer factory")
            self.x[k] = v

    def new(self, path=None, *args, **kwargs):
        writer_keyed = None
        writer_immutable = None
        if self.key_writer_constructor is not None:
            writer_keyed = self.key_writer_constructor(path, **self.x)
        if self.immutable_writer_constructor is not None:
            writer_immutable = self.immutable_writer_constructor(path, **self.x)
        return KeyedWriter(writer_keyed, writer_immutable)


class MetadataWriter(OutputWriter):
    """Custom writer for publishing data under immutable content-addressed pointers in the cic-meta storage backend.

    Data that is not utf-8 will be converted to base64 before publishing.

    Implements cic.writers.OutputWriter
    """

    def write(self, k, v):
        rq = MetadataRequestsHandler(MetadataPointer.NONE, bytes.fromhex(k))
        try:
            v = v.decode("utf-8")
            v = json.loads(v)
            logg.debug(f"metadatawriter bindecode {k} {v}")
        except UnicodeDecodeError:
            v = base64.b64encode(v).decode("utf-8")
            v = json.loads(json.dumps(v, separators=(",", ":")))
            logg.debug(f"metadatawriter b64encode {k} {v}")
        r = rq.create(v)
        logg.info(f"metadata submitted at {k}")
        return r



# standard imports
import os
import logging
import sys
import importlib

# external imports
import chainlib.cli

# local imports
import cic.cmd.init as cmd_init
import cic.cmd.show as cmd_show
import cic.cmd.ext as cmd_ext
import cic.cmd.export as cmd_export
import cic.cmd.wizard as cmd_wizard
from cic.config import ensure_base_configs

logging.basicConfig(level=logging.WARNING)
logg = logging.getLogger()

script_dir = os.path.dirname(os.path.realpath(__file__))
data_dir = os.path.join(script_dir, '..', 'data')
base_config_dir = os.path.join(data_dir, 'config')
schema_dir = os.path.join(script_dir, '..', 'schema')
user_config_dir = os.path.join(os.path.expanduser("~"), ".config", "cic", "cli", "config") 

arg_flags = chainlib.cli.argflag_std_read | chainlib.cli.Flag.SEQ
argparser = chainlib.cli.ArgumentParser(
    env=os.environ,
    arg_flags=arg_flags,
    description='CIC cli tool for generating and publishing tokens'
)

sub = argparser.add_subparsers()
sub.dest = 'command'

sub_init = sub.add_parser('init', help='initialize new cic data directory')
cmd_init.process_args(sub_init)

sub_show = sub.add_parser('show', help='display summary of current state of cic data directory')
cmd_show.process_args(sub_show)

sub_export = sub.add_parser('export', help='export cic data directory state to a specified target')
cmd_export.process_args(sub_export)

sub_ext = sub.add_parser('ext', help='extension helpers')
cmd_ext.process_args(sub_ext)

sub_wizard = sub.add_parser('wizard', help='An interactive wizard for creating and publishing contracts')
cmd_wizard.process_args(sub_wizard)

args = argparser.parse_args(sys.argv[1:])

if args.command is None:
    logg.critical('Subcommand missing')
    sys.stderr.write("\033[;91m" + 'subcommand missing' + "\033[;39m\n")
    argparser.print_help(sys.stderr)
    sys.exit(1)

modname = f'cic.cmd.{args.command}'
logg.debug(f'using module {modname}')
cmd_mod = importlib.import_module(modname)

extra_args = {
    'p': 'RPC_PROVIDER',
}
ensure_base_configs(user_config_dir)

config = chainlib.cli.Config.from_args(args, arg_flags=arg_flags, base_config_dir=base_config_dir, extra_args=extra_args)

def main():
    try:
        cmd_mod.execute(config, args)
    except Exception as e:
        logg.exception(e)
        sys.stderr.write("\033[;91m" + str(e) + "\033[;39m\n")
        sys.exit(1)

if __name__ == '__main__':
    main()



# standard imports
import logging
import os

# local imports
from cic.contract.base import Data, data_dir

logg = logging.getLogger(__name__)


class Attachment(Data):
    """Processes, serialized and publishes all attachments found in the "attachments" subdirectory of the settings directory.

    :param path: Path to settings directory
    :type path: str
    :param writer: Writer interface receiving the output of the processor
    :type writer: cic.writers.OutputWriter
    """

    def __init__(self, path=".", writer=None, interactive=False):
        super(Attachment, self).__init__()
        self.contents = {}
        self.path = path
        self.writer = writer
        self.attachment_path = os.path.join(self.path, "attachments")
        if interactive:
            self.start()
            input(
                f"Please add attachment files to '{os.path.abspath(os.path.join(self.path,'attachments'))}' and then press ENTER to continue"
            )
            self.load()

    def load(self):
        """Loads attachment data from settings."""
        for s in os.listdir(self.attachment_path):
            fp = os.path.realpath(os.path.join(self.attachment_path, s))
            with open(fp, "rb") as f:
                r = f.read()

            z = self.hash(r).hex()
            self.contents[z] = fp

            logg.debug(f"loaded attachment file {fp} digest {z}")

    def start(self):
        """Initialize attachment settings from template."""
        super(Attachment, self).start()
        os.makedirs(self.attachment_path)

    def get(self, k):
        """Get a single attachment by the sha256 hash of the content.

        :param k: Content hash
        :type k: str (hex)
        """
        return self.contents[k]

    def asdict(self):
        """Output attachment state to dict"""
        return self.contents

    def process(self, token_address=None, token_symbol=None, writer=None):
        """Serialize and publish attachments.

        See cic.processor.Processor.process
        """
        if writer == None:
            writer = self.writer

        for key, value in self.contents.items():
            fp = os.path.join(self.attachment_path, value)
            with open(fp, "rb") as f:
                data = f.read()
            logg.debug(f"writing attachment {key}")
            writer.write(key, data)

    def __str__(self):
        s = ""
        for key, value in self.contents.items():
            s += f"{key} = {value}\n"  # self.digests[i].hex(), self.contents[i])

        return s
